# from esxi_utils.vm.types.ostype import OSType
# from esxi_utils.vm.types.cisco import CiscoVirtualMachine
# from esxi_utils.vm.types.linux import LinuxVirtualMachine
# from esxi_utils.vm.types.panos import PaloAltoFirewallVirtualMachine
# from esxi_utils.vm.types.windows import WindowsVirtualMachine
